/*
 * Copyright (c) 2010 Sendmail, Inc. and its suppliers.
 *	All rights reserved.
 *
 * By using this file, you agree to the terms and conditions set
 * forth in the LICENSE file which can be found at the top level of
 * the sendmail distribution.
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: t-dkim-1.c,v 1.28 2013/12/03 02:52:52 ca Exp $")

#include "sm/assert.h"
#include "sm/error.h"
#include "sm/stat.h"
#include "sm/fcntl.h"
#include "sm/ctype.h"
#include "sm/io.h"
#include "sm/str.h"
#include "sm/string.h"
#include "sm/mta.h"
#include "sm/misc.h"
#include "sm/base64.h"
#include "sm/dkimdef.h"
#include "sm/dkim.h"
#include "sm/sysexits.h"
#include "sm/units.h"
#include "sm/misc.h"
#include "sm/test.h"

#include "dkim.h"

#if MTA_USE_DKIM
#define SIG "v=1; a=rsa-sha256; c=simple/simple; d=example.net;\r\n\ts=default; t=1258487282;\r\n\tbh=P7M5AsnZKX0CSpQl3w3u9XwzYQ7gDrSE5gSSQy/1W2E=; h=From:To:Subject;\r\n\tb=bJ2PgR07u9NdHD8UgExfjc6UGg0PD8Ac7yg1YlSlWFSW1oboC4jUnzBTsCFE6Gqml\r\n\t yQISCypKClJt1746n3UYa61bOkcTiEeD27AIX+bbEXWqyw/+Y9RNHehcYT7eY1ITG3\r\n\t C1xeoJ2aNeiEPcG2SPYf8JsuY8dzzeOGe7CONvGs="

#define BH "P7M5AsnZKX0CSpQl3w3u9XwzYQ7gDrSE5gSSQy/1W2E="

static const char *bodyhash64 = BH;
static const char *sig64 = NULL;

#define BODY "2020 Vision Street\r\nAnycity, Anystate 90009\r\nOffice ph: (510) 555-1234\r\nCell ph:   (510) 555-5678\r\nfax:       (510) 555-9000\r\nemail: mmtest3@v253.pogonip.net\r\n"

static uint Verbose = 0;
dkim_s_T dkim_s;

/*
**  DKIM_TEST -- simple DKIM test
**
**	Parameters:
**		msg -- msg
**		len -- len of msg
**
**	Returns:
**		usual sm_error code
*/

static sm_ret_T
dkim_test(dkim_ctx_P dkim_ctx, const uchar *msg, size_t len, ssize_t chunksize, time_T dkim_time)
{
	sm_ret_T ret;
	dkim_msg_P dkim_msg;

	ret = SM_E_UNEXPECTED;
	SM_TEST_E(msg != NULL);
	SM_TEST_E(*msg != '\0');
	SM_TEST_E(len != 0);

	ret = dkim_msg_open(dkim_ctx, &dkim_msg);
	SM_TEST_E(SM_SUCCESS == ret);

	dkim_msg->dkm_s = &dkim_s;

	dkim_msg->dkm_time = dkim_time;

	if (chunksize < 0) {
		ret = dkim_msg_chunk(dkim_msg, msg, len);
		SM_TEST_E(SM_SUCCESS == ret);
	}
	else {
		size_t offset;
		bool last;

		for (offset = 0, last = false; !last; offset += chunksize) {
			if (chunksize + offset > len) {
				chunksize = len - offset;
				last = true;
			}
			ret = dkim_msg_chunk(dkim_msg, msg + offset, chunksize);
			if (Verbose > 2)
				sm_io_fprintf(smioerr,
					"dkim_msg=%x, offset=%u, chunksize=%ld\n",
					ret, offset, chunksize);
			if (last)
				SM_TEST_E(SM_SUCCESS == ret);
			else
				SM_TEST_E(SM_NEEDMORE == ret);
		}
	}

	if (bodyhash64 != NULL)
		SM_TEST(strstr((const char *)sm_str_getdata(dkim_msg->dkm_sighdr), bodyhash64) != NULL);
	if (sig64 != NULL) {
		uint u;
		sm_str_P sig;

		sig = sm_str_new(NULL, 256, 4096);	/* length??? */
		SM_TEST_E(sig != NULL);
		for (u = 0; u < sm_str_getlen(dkim_msg->dkm_sighdr); u++) {
			uchar c;

			c = sm_str_rd_elem(dkim_msg->dkm_sighdr, u);
			if (ISSPACE(c))
				continue;
			ret = sm_str_put(sig, c);
			SM_TEST_E(SM_SUCCESS == ret);
		}
		SM_TEST(strstr((const char *)sm_str_getdata(sig), sig64) != NULL);
	}
	if (Verbose > 0)
		sm_io_fprintf(smioerr, "sig: %S\n", dkim_msg->dkm_sighdr);

	ret = dkim_msg_close(dkim_msg);
	SM_TEST_E(SM_SUCCESS == ret);
	return ret;

  error:
	return ret;
}

/*
**  USAGE -- usage message
**
**	Parameters:
**		prg -- program name
**
**	Returns:
**		none
*/

static void
usage(char *prg)
{
	sm_io_fprintf(smioerr, "usage: %s [options]\n", prg);
	sm_io_fprintf(smioerr,
		"-b text     use text as msg\n"
		"-c n        chunksize (instead of entire file at once)\n"
		"-d domain   use domain for signing\n"
		"-h hash     expected base64 bodyhash\n"
		"-k file     file contains private key for DKIM signing\n"
		"-m file     read file for msg\n"
		"-R          relaxed body canonicalization\n"
		"-r          relaxed header canonicalization\n"
		"-S hash     expected base64 signature\n"
		"-s selector use selector for signing\n"
		"-t time     use time for signing\n"
		"-V          increase verbosity\n"
		);
	return;
}
#endif /* MTA_USE_DKIM */

int
main(int argc, char *argv[])
{
#if MTA_USE_DKIM
	int c, r;
	size_t msglen;
	ssize_t chunksize;
	uchar *msg;
	dkim_ctx_T dkim_ctx;
	dkim_cnf_T dkim_cnf;
	time_T dkim_time;

	msg = NULL;

	/* fixed time (magic value), see SIG */
	dkim_time = 1258487282;
	msglen = 0;
	sm_memzero(&dkim_ctx, sizeof(dkim_ctx));
	sm_memzero(&dkim_cnf, sizeof(dkim_cnf));
	sm_memzero(&dkim_s, sizeof(dkim_s));
	dkim_ctx.dkc_cnf = &dkim_cnf;
	chunksize = -1;
	dkim_s.dks_domain = "example.net";
	dkim_s.dks_selector = "default";
	dkim_cnf.dkimcnf_hdrtlen = SM_TOTAL_HDR_LEN;
	dkim_cnf.dkimcnf_hdrvlen = SM_HDR_LEN_MAX;

	while ((c = getopt(argc, argv, "b:Cc:d:h:k:m:pRrS:s:t:V")) != -1) {
		switch (c) {
		  case 'b':
			msg = (uchar *)strdup(optarg);
			msglen = strlen((const char *)msg);
			break;
		  case 'C':
			compoptions(smioout);
			return 0;
		  case 'c':
			chunksize = strtol(optarg, NULL, 0);
			break;
		  case 'd':
			dkim_s.dks_domain = strdup(optarg);
			break;
		  case 'k':
			dkim_s.dks_keyfile = optarg;
			r = dkim_readkey(&dkim_s);
			if (r != 0)
				return r;
			break;
		  case 'h':
			bodyhash64 = strdup(optarg);
			break;
		  case 'm':
			r = readfile(optarg, &msg, &msglen);
			if (r != 0)
				return r;
			break;
		  case 'R':
			dkim_cnf.dkimcnf_bcanon = DKIM_CANON_RELAXED;
			break;
		  case 'r':
			dkim_cnf.dkimcnf_hcanon = DKIM_CANON_RELAXED;
			break;
		  case 'S':
			sig64 = strdup(optarg);
			break;
		  case 's':
			dkim_s.dks_selector = strdup(optarg);
			break;
		  case 't':
			dkim_time = strtol(optarg, NULL, 0);
			break;
		  case 'V':
			++Verbose;
			break;
		  default:
			usage(argv[0]);
			return EX_USAGE;
		}
	}

	sm_test_begin(argc, argv, "test dkim 1");
	dkim_test(&dkim_ctx, msg, msglen, chunksize, dkim_time);
	return sm_test_end();
#else
	sm_io_fprintf(smioerr, "%s: SKIPPED: MTA_USE_DKIM not set\n", argv[0]);
	return 0;
#endif
}
