/*
 * Copyright (c) 2004, 2011 Sendmail, Inc. and its suppliers.
 *	All rights reserved.
 *
 * By using this file, you agree to the terms and conditions set
 * forth in the LICENSE file which can be found at the top level of
 * the sendmail distribution.
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: t-restart.c,v 1.13 2014/01/06 20:20:30 ca Exp $")

#include "sm/error.h"
#include "sm/ctype.h"
#include "sm/sysexits.h"
#include "sm/signal.h"
#include "sm/stat.h"
#include "sm/wait.h"

#include <stdio.h>

#define PIDFN "t-restart.pid"
#define SLEEPT	60

static int Verbose = 0;
static int termsleep = 0;

#include "rdwrcounter.c"

struct e2s_S
{
	int			 e2s_code;
	const char	*e2s_str;
};
typedef struct e2s_S	e2s_T, *e2s_P;

static e2s_T
ex2txt[] = {
#ifdef EX_OK
	{EX_OK, "EX_OK" },
#endif 
#ifdef EX__BASE
	{EX__BASE, "EX__BASE" },
#endif 
#ifdef EX_USAGE
	{EX_USAGE, "EX_USAGE" },
#endif 
#ifdef EX_DATAERR
	{EX_DATAERR, "EX_DATAERR" },
#endif 
#ifdef EX_NOINPUT
	{EX_NOINPUT, "EX_NOINPUT" },
#endif 
#ifdef EX_NOUSER
	{EX_NOUSER, "EX_NOUSER" },
#endif 
#ifdef EX_NOHOST
	{EX_NOHOST, "EX_NOHOST" },
#endif 
#ifdef EX_UNAVAILABLE
	{EX_UNAVAILABLE, "EX_UNAVAILABLE" },
#endif 
#ifdef EX_SOFTWARE
	{EX_SOFTWARE, "EX_SOFTWARE" },
#endif 
#ifdef EX_OSERR
	{EX_OSERR, "EX_OSERR" },
#endif 
#ifdef EX_OSFILE
	{EX_OSFILE, "EX_OSFILE" },
#endif 
#ifdef EX_CANTCREAT
	{EX_CANTCREAT, "EX_CANTCREAT" },
#endif 
#ifdef EX_IOERR
	{EX_IOERR, "EX_IOERR" },
#endif 
#ifdef EX_TEMPFAIL
	{EX_TEMPFAIL, "EX_TEMPFAIL" },
#endif 
#ifdef EX_PROTOCOL
	{EX_PROTOCOL, "EX_PROTOCOL" },
#endif 
#ifdef EX_NOPERM
	{EX_NOPERM, "EX_NOPERM" },
#endif 
#ifdef EX_CONFIG
	{EX_CONFIG, "EX_CONFIG" },
#endif 
#ifdef EX__MAX
	{EX__MAX, "EX__MAX" },
#endif 
#ifdef EX_RESTARTDEP
	{EX_RESTARTDEP, "EX_RESTARTDEP" },
#endif 
#ifdef EX_RESTARTALL
	{EX_RESTARTALL, "EX_RESTARTALL" },
#endif 
#ifdef EX_TRYLATER
	{EX_TRYLATER, "EX_TRYLATER" },
#endif 
	{0, NULL }
};

static int
findt(const char *text, e2s_P table)
{
	int i;

	i = 0;
	while (table[i].e2s_code != 0 || table[i].e2s_str != NULL) {
		if (strcmp(table[i].e2s_str, text) == 0)
			return table[i].e2s_code;
		++i;
	}
	return -1;
}


static void
usage(char *prg, char c)
{
	fprintf(stderr,
		"usage: %s [options] -f file\n"
		"options:\n"
		"-A          use exit code EX_RESTARTALL\n"
		"-C file     create file when starting, remove at end\n"
		"-c n        catch sigterm, sleep n seconds before _exit()\n"
		"-f file     file for counter\n"
		"-L          use exit code EX_TRYLATER\n"
		"-n n        use n as limit [0]\n"
		"-o          use exit code only for first invocation\n"
		"-P          use exit code EX_NOPERM\n"
		"-p pidfile  use pidfile for writing pid [%s]\n"
		"-r n        use exit code n [0]\n"
		"-R          use exit code EX_RESTARTDEP\n"
		"-S file     check whether file exists (once per second)\n"
		"-s n        sleep n seconds [%d]\n"
		"-T          use exit code EX_TEMPFAIL\n"
		"-U          use exit code EX_CONFIG\n"
		"-V          increase verbosity\n"
		"-W          use exit code EX_SOFTWARE\n"
		"-X EX_code  use exit code EX_code\n"
		"\n%s reads a counter from file and writes back counter + 1\n"
		"if counter is less than the limit then the program exits "
		"immediately\notherwise it will sleep for a while\n"
		, prg, PIDFN, SLEEPT, prg);
	if (ISPRINT(c))
		fprintf(stderr, "usage: %s invalid option %c\n", prg, c);
	exit(EX_USAGE);
}

#define	SIGTERMMSG		"Got SIGTERM\n"

static void
sighdlr(int sig)
{
	if (sig == SIGTERM) {
		write(STDERR_FILENO, SIGTERMMSG, sizeof(SIGTERMMSG));
		if (termsleep > 0)
			sleep(termsleep);
		_exit(SIGTERM);
	}
}


static int
wrtfile(const char *prg, const char *fn, pid_t pid)
{
	FILE *pidfp;

	/* write pid */
	pidfp = fopen(fn, "w");
	if (NULL == pidfp) {
		fprintf(stderr, "%s: can't open \"%s\"\n", prg, fn);
		exit(EX_OSERR);
	}
	fprintf(pidfp, "%ld\n", (long) pid);
	fclose(pidfp);
	pidfp = NULL;
	return 0;
}

int
main(int argc, char *argv[])
{
	int abort, c, cnt, cur, sleept, rc, rconce, waitt;
	pid_t pid;
	char *cntfn, *pidfn, *prg, *server, *client;

	prg = argv[0];
	cntfn = NULL;
	pidfn = PIDFN;
	pid = getpid();
	sleept = SLEEPT;
	server = NULL;
	client = NULL;
	rc = cnt = 0;
	rconce = 0;
	abort = 0;
	waitt = 4;
	while ((c = getopt(argc, argv, "AaC:c:f:Ln:oPp:Rr:S:s:TUVWw:X:x:")) != -1) {
		switch (c) {
		  case 'A':
			rc = EX_RESTARTALL;
			break;
		  case 'a':
			abort = 1;
			break;
		  case 'C':
			client = optarg;
			break;
		  case 'c':
			termsleep = atoi(optarg);
			signal(SIGTERM, sighdlr);
			break;
		  case 'f':
			cntfn = optarg;
			break;
		  case 'L':
			rc = EX_TRYLATER;
			break;
		  case 'n':
			cnt = atoi(optarg);
			break;
		  case 'o':
			rconce = 1;
			break;
		  case 'P':
			rc = EX_NOPERM;
			break;
		  case 'p':
			pidfn = optarg;
			break;
		  case 'R':
			rc = EX_RESTARTDEP;
			break;
		  case 'r':
			rc = atoi(optarg);
			break;
		  case 'S':
			server = optarg;
			break;
		  case 's':
			sleept = atoi(optarg);
			break;
		  case 'T':
			rc = EX_TEMPFAIL;
			break;
		  case 'U':
			rc = EX_CONFIG;
			break;
		  case 'V':
			++Verbose;
			break;
		  case 'W':
			rc = EX_SOFTWARE;
			break;
		  case 'w':
			waitt = atoi(optarg);
			break;
		  case 'X':
			rc = findt(optarg, ex2txt);
			break;
		  case 'x':
			rc = strtoul(optarg, NULL, 0);
			break;
		  case 'h':
		  case '?':
		  default:
			usage(prg, c);
			break;
		}
	}

	if (Verbose > 0)
		fprintf(stderr, "%s: pid=%ld, where=start\n", prg, (long) pid);

	/* get current invocation counter */
	if (NULL == cntfn) {
		fprintf(stderr, "%s: missing -f filename\n", prg);
		return EX_USAGE;
	}
	c = rdwrcounter(prg, cntfn, &cur);
	if (c != 0)
		return c;

	if (Verbose > 0)
		fprintf(stderr, "%s: pid=%ld, cur=%d\n", prg, (long) pid, cur);

	/* ugly hack: reset to "default" values after first invocation */
	if (rc != 0 && cur > 0 && rconce) {
		rc = 0;
		sleept = SLEEPT;
	}

	wrtfile(prg, pidfn, pid);

	if (Verbose > 0)
		fprintf(stderr, "%s: pid=%ld, cur=%d, n=%d\n", prg, (long) pid, cur, cnt);

	/* something went wrong, it shouldn't go that far? */
	if (cur >= cnt)
		return EX_CONFIG;

	if (client != NULL)
		wrtfile(prg, client, pid);
	else if (server != NULL) {
		int sl, exists;
		struct stat sb;

		sl = 1;
		sleep(1);
		while ((exists = stat(server, &sb)) != 0) {
			sleep(1);
			++sl;
			if (sl >= waitt)
				break;
		}
		if (exists != 0)
			rc = EX_RESTARTDEP;
	}
	else
		sleep(sleept);

	if (Verbose > 0) {
		const char *status;

		status = ex2txt_r(rc);
		fprintf(stderr, "%s: pid=%ld, done=%d, status=%s\n", prg, (long) pid, rc, status != NULL ? status : "unknown");
	}
	if (client != NULL) {
		c = unlink(client);
		if (c != 0) {
			fprintf(stderr, "%s: can't unlink \"%s\"\n", prg, client);
			exit(EX_OSERR);
		}
	}
	assert(abort == 0);

	return rc;
}
